import random
import csv
from deap import base
from deap import creator
from deap import tools
from deap import algorithms
import pandas as pd
import numpy as np
import QSTK.qstkfeat.featutil as ftu
import QSTK.qstkutil.qsdateutil as du
import datetime as dt
import QSTK.qstkutil.DataAccess as da
from pyalgotrade import strategy
from pyalgotrade import barfeed
from pyalgotrade import bar
from pyalgotrade.technical import ma as pytechma
from pyalgotrade import plotter
from pyalgotrade.stratanalyzer import returns
from pyalgotrade.stratanalyzer import sharpe
from pyalgotrade.utils import stats as pytechstats

# stocks used as training data

stock = [
    'AAPL',
    'ADSK',
    'AKAM',
    'AMZN',
    'CSCO',
    'EA',
    'INTU',
    'NFLX',
    'NVDA',
    'QCOM',
    'MSFT',
    'IBM',
    ]
start = dt.datetime(2010, 1, 1)
end = dt.datetime(2012, 1, 1)

# data import

timeofday = dt.timedelta(hours=16)
timestamps = du.getNYSEdays(start, end, timeofday)
dataobj = da.DataAccess('Yahoo')
keys = [
    'open',
    'high',
    'low',
    'close',
    'volume',
    'actual_close',
    ]
data = dataobj.get_data(timestamps, stock, keys)
dataframe = dict(zip(keys, data))
for key in keys:
    dataframe[key] = dataframe[key].fillna(method='ffill')
    dataframe[key] = dataframe[key].fillna(method='bfill')
    dataframe[key] = dataframe[key].fillna(1.0)

# data import to verify the model

verify_date_start = end
verify_date_end = dt.datetime(2013, 1, 1)
verify_date_timestamps = du.getNYSEdays(verify_date_start,
        verify_date_end, timeofday)
verify_data = dataobj.get_data(verify_date_timestamps, stock, keys)
verify_dataframe = dict(zip(keys, verify_data))
for key in keys:
    verify_dataframe[key] = verify_dataframe[key].fillna(method='ffill')
    verify_dataframe[key] = verify_dataframe[key].fillna(method='bfill')
    verify_dataframe[key] = verify_dataframe[key].fillna(1.0)

# calculated rsi for training dataframe

rsi = ftu.featRSI(dataframe).fillna(0.0)

# calculated moving average for training dataframe

ma = ftu.featMA(dataframe, bRel=False).fillna(0.0)

# calculated momentum for training dataframe

mom = ftu.featMomentum(dataframe).fillna(0.0)

# calculated stochastic oscilator for training dataframe

stoch = ftu.featStochastic(dataframe).fillna(0.0)

# calculating indicators for the verifying dataframe here
# calculated rsi for verifying dataframe

verify_rsi = ftu.featRSI(verify_dataframe).fillna(0.0)

# calculated moving average for verifying dataframe

verify_ma = ftu.featMA(verify_dataframe, bRel=False).fillna(0.0)

# calculated momentum for verifying dataframe

verify_mom = ftu.featMomentum(verify_dataframe).fillna(0.0)

# calculated stochastic oscilator for verifying dataframe

verify_stoch = ftu.featStochastic(verify_dataframe).fillna(0.0)

toolbox = base.Toolbox()


# basic initialization for deap

def initDeap():
    creator.create('FitnessMin', base.Fitness, weights=(-1.0, ))
    creator.create('Individual', list, fitness=creator.FitnessMin)
    IND_SIZE = 4

    # registerd coeff alias to coefficient function

    toolbox.register('coeffs', coefficient)

    # individual alias registerd to tools.initRepeat

    toolbox.register('individual', tools.initRepeat,
                     creator.Individual, toolbox.coeffs, n=IND_SIZE)

    # population alias registered

    toolbox.register('population', tools.initRepeat, list,
                     toolbox.individual)

    # evaluate fucntion registered to computeFitness function

    toolbox.register('evaluate', computeFitness)
    toolbox.register('mate', tools.cxSimulatedBinary, eta=0.3)
    toolbox.register('mutate', tools.mutGaussian, mu=0.0, sigma=1.0,
                     indpb=0.05)
    toolbox.register('select', tools.selTournament, tournsize=5)


# 1. rsi, 2. ma, 3. mom, 4. stock
# above is the order of coefficients
# removed pivot point as one of the techincal indicator, because lets say we are running this for the next 6 months,
# in that case we would not have the value of high, low of future. as such for future we can't use it.
# also saw from many simulations, the coefficient of pivot point was usually very low

def computeFitness(individual):
    error = 0.0
    count = 0
    for key in stock:

        # difference in RSI between day1 and day1 + 1

        rsidiff = rsi[key] - rsi[key].shift()

        # difference in moving average between day1 and day1 + 1

        madiff = ma[key] - ma[key].shift()

        # difference in momentum between day1 and day1 + 1

        momdiff = mom[key] - mom[key].shift()
        per_stock_count = 0
        dataframe_index = dataframe['close'][key].index.tolist()
        for (date, val) in rsi[key].iteritems():
            next_day = (dataframe_index[per_stock_count
                        + 1] if per_stock_count <= len(dataframe_index)
                        - 1 else None)
            if not np.isnan(rsidiff[date]) and next_day is not None:

                #            Fitness defined as
                # market close (day2) - [ rsi difference between consecutive day * (coeff 1) +
                # moving average difference between consecutive day * (coeff 2) +
                # momentum difference between consecutive day * (coeff 3) +
                # stochastic oscillator (itself a kind of predictor) * (coeff 4)
                # Objective is to minimize the summation of squares of the above differences

                ft = rsidiff[date] * individual[0] + madiff[date] \
                    * individual[1] + momdiff[date] * individual[2] \
                    + stoch[key][date] * individual[3]
                ft = ft + dataframe['close'][key][date]
                error = error + np.square(dataframe['close'
                        ][key][next_day] - ft)
                count = count + 1
                per_stock_count = per_stock_count + 1

    # dividing by total count gives the mean square error, which we want to minimize

    return (error / count, )


def coefficient():
    return random.random()


def predict(individual):
    error = 0.0
    count = 0
    prediction_series = []

    # using the same logic as fitness function to calculate the predicted values

    for key in stock:

        # difference in RSI between day1 and day1 + 1

        rsidiff = verify_rsi[key] - verify_rsi[key].shift()

        # difference in moving average between day1 and day1 + 1

        madiff = verify_ma[key] - verify_ma[key].shift()

        # difference in momentum between day1 and day1 + 1

        momdiff = verify_mom[key] - verify_mom[key].shift()
        per_stock_count = 0
        verify_dataframe_indexes = verify_dataframe['close'
                ][key].index.tolist()
        for (date, val) in verify_rsi[key].iteritems():
            next_day = (verify_dataframe_indexes[per_stock_count
                        + 1] if per_stock_count
                        <= len(verify_dataframe_indexes) - 1 else None)
            if not np.isnan(rsidiff[date]) and next_day is not None:

                ft = rsidiff[date] * individual[0] + madiff[date] \
                    * individual[1] + momdiff[date] * individual[2] \
                    + verify_stoch[key][date] * individual[3]
                ft = ft + verify_dataframe['close'][key][date]
                prediction_series.append((key, next_day,
                        verify_dataframe['close'][key][next_day], ft,
                        verify_dataframe['close'][key][next_day] - ft))
                error = error + np.square(verify_dataframe['close'
                        ][key][next_day] - ft)

                # replacing the close price with the one predicted

                verify_dataframe['close'][key][next_day] = ft
                count = count + 1
                per_stock_count = per_stock_count + 1

    # dividing by total count gives the mean square error, which we wanted to minimize

    return (error / count, prediction_series)


if __name__ == '__main__':
    initDeap()

    # population generated

    pop = toolbox.population(n=50)
    hof = tools.HallOfFame(1)
    stats = tools.Statistics(lambda ind: ind.fitness.values)
    stats.register('avg', np.mean)
    stats.register('min', np.min)
    stats.register('max', np.max)
    stats.register('std', np.std)
    stats.register('median', np.median)

    # built in algo used with cross over probability = 0.5, mutation probability = 0.2 and number of generation = 5

    (pop, logbook) = algorithms.eaSimple(
        pop,
        toolbox,
        cxpb=0.5,
        mutpb=0.2,
        ngen=5,
        stats=stats,
        halloffame=hof,
        verbose=True,
        )

    print 'Best individual is: %s\nwith fitness: %s' % (hof[0],
            hof[0].fitness)


#  class defining our trading strategy

class TradingStrategy(strategy.BacktestingStrategy):

    # default period to calculate moving avg is 14 , cash = 100000

    def __init__(
        self,
        barFeed,
        instrument,
        smaPeriod=26,
        cash_or_brk=100000,
        ):
        strategy.BacktestingStrategy.__init__(self, barFeed,
                cash_or_brk)
        self.__position = None
        self.__instrument = instrument
        self.__sma = \
            pytechma.SMA(barFeed[instrument].getPriceDataSeries(),
                         smaPeriod)

    # for every day , we decide using this function if we need to buy or sell.
    # currently we simply see if the closing price > moving average, we buy, else we sell

    def onBars(self, bars):

        # Wait for enough bars to be available to calculate a SMA.

        if self.__sma[-1] is None:
            return

        bar = bars[self.__instrument]

        # If a position was not opened, check if we should enter a long position.

        if self.__position is None:
            if bar.getPrice() > self.__sma[-1]:

                # Enter a buy market order for 5 shares. The order is good till canceled.

                self.__position = self.enterLong(self.__instrument, 5,
                        True)
        elif bar.getPrice() < self.__sma[-1] \
            and not self.__position.exitActive():

        # Check if we have to exit the position.

            self.__position.exitMarket()

    # this function are called when the order submitted to enter a position was filled

    def onEnterOk(self, position):
        execInfo = position.getEntryOrder().getExecutionInfo()
        self.info('BUY %s at $%.2f' % (self.__instrument,
                  execInfo.getPrice()))

    def onEnterCanceled(self, position):
        self.__position = None

    # this function are called when the exit order for a position was filled

    def onExitOk(self, position):
        execInfo = position.getExitOrder().getExecutionInfo()
        self.info('SELL %s at $%.2f' % (self.__instrument,
                  execInfo.getPrice()))
        self.__position = None

    def onExitCanceled(self, position):

        # If the exit was canceled, re-submit it.

        self.__position.exitMarket()

    # if we are using other indicators as well, then we can define fcuntions to expose the data series here

    def getSMA(self):
        return self.__sma


# class representing bar

class DataframeBar(bar.Bar):

    def __init__(
        self,
        dateTime,
        open_,
        high,
        low,
        close,
        volume,
        adjClose,
        frequency,
        ):
        self.__dateTime = dateTime
        self.__open = open_
        self.__close = close
        self.__high = high
        self.__low = low
        self.__volume = volume
        self.__adjClose = adjClose
        self.__frequency = frequency
        self.__useAdjustedValue = False

    def setUseAdjustedValue(self, useAdjusted):
        if useAdjusted and self.__adjClose is None:
            raise Exception('Adjusted close is not available')
        self.__useAdjustedValue = useAdjusted

    def getUseAdjValue(self):
        return self.__useAdjustedValue

    def getDateTime(self):
        return self.__dateTime

    def getOpen(self, adjusted=False):
        if adjusted:
            if self.__adjClose is None:
                raise Exception('Adjusted close is missing')
            return self.__adjClose * self.__open / float(self.__close)
        else:
            return self.__open

    def getHigh(self, adjusted=False):
        if adjusted:
            if self.__adjClose is None:
                raise Exception('Adjusted close is missing')
            return self.__adjClose * self.__high / float(self.__close)
        else:
            return self.__high

    def getLow(self, adjusted=False):
        if adjusted:
            if self.__adjClose is None:
                raise Exception('Adjusted close is missing')
            return self.__adjClose * self.__low / float(self.__close)
        else:
            return self.__low

    def getClose(self, adjusted=False):
        if adjusted:
            if self.__adjClose is None:
                raise Exception('Adjusted close is missing')
            return self.__adjClose
        else:
            return self.__close

    # These are helper functions to get the prices

    def getVolume(self):
        return self.__volume

    def getAdjOpen(self):
        return self.getOpen(True)

    def getAdjHigh(self):
        return self.getHigh(True)

    def getAdjLow(self):
        return self.getLow(True)

    def getAdjClose(self):
        return self.__adjClose

    def getFrequency(self):
        return self.__frequency

    def getPrice(self):
        if self.__useAdjustedValue:
            return self.__adjClose
        else:
            return self.__close


# class representing a barfeed, created from a panda dataframe

class DataFrameFeed(barfeed.BaseBarFeed):

    def __init__(
        self,
        dataframe,
        instrument,
        frequency,
        ):
        super(DataFrameFeed, self).__init__(frequency)
        self.registerInstrument(instrument)
        self.__df = dataframe
        self.__instrument = instrument
        self.__next = 0

    def reset(self):
        super(DataFrameFeed, self).reset()
        self.__next = 0

    def peekDateTime(self):
        return self.getCurrentDateTime()

    def getCurrentDateTime(self):
        ret = None
        if not self.eof():
            rowkey = self.__df.index.tolist()[self.__next]
            ret = rowkey.to_datetime()
        return ret

    def barsHaveAdjClose(self):
        return True

    def getNextBars(self):
        ret = None
        if not self.eof():
            rowkey = self.__df.index.tolist()[self.__next]
            row = self.__df.ix[rowkey]
            bar_dict = {self.__instrument: DataframeBar(
                rowkey.to_datetime(),
                row[keys[0]],
                row[keys[1]],
                row[keys[2]],
                row[keys[3]],
                row[keys[4]],
                row[keys[5]],
                self.getFrequency(),
                )}
            ret = bar.Bars(bar_dict)
            self.__next += 1
        return ret

    def eof(self):
        return self.__next >= len(self.__df.index)

    def start(self):
        pass

    def stop(self):
        pass

    def join(self):
        pass


# class defining % returns for cumulative and net

class PercentageReturns(returns.Returns):

    def __init__(self):
        super(PercentageReturns, self).__init__()

    def beforeAttach(self, strat):

        # Get or create a shared ReturnsAnalyzerBase

        analyzer = returns.ReturnsAnalyzerBase.getOrCreateShared(strat)
        analyzer.getEvent().subscribe(self.__onReturns)

    def __onReturns(self, dateTime, returnsAnalyzerBase):

        # multiplying by 100 gives us the percentage, weird I know, but the examples of pyalgotrade seems to say this

        super(PercentageReturns,
              self).getReturns().append(returnsAnalyzerBase.getNetReturn()
                * 100)
        super(PercentageReturns,
              self).getCumulativeReturns().append(returnsAnalyzerBase.getCumulativeReturn()
                * 100)


# for each stocks we are running our strategy

total_portfolio_value = 0
for key in stock:
    data_stock = {
        'open': verify_dataframe['open'][key],
        'high': verify_dataframe['high'][key],
        'low': verify_dataframe['low'][key],
        'close': verify_dataframe['close'][key],
        'volume': verify_dataframe['volume'][key],
        'actual_close': verify_dataframe['actual_close'][key],
        }

    # creating a dataframe as expected by dataframefeed

    df = pd.DataFrame(data_stock)  # create a dataframe feed with frequncy daily
    feed = DataFrameFeed(df, key, barfeed.Frequency.DAY)

    # the startegy for each stock with deafult cash_or_brk = 100000

    investing = 100000
    tstrategy = TradingStrategy(feed, key, cash_or_brk=investing)
    returnsAnalyzer = PercentageReturns()
    tstrategy.attachAnalyzer(returnsAnalyzer)

    sharpeRatioAnalyzer = sharpe.SharpeRatio()
    tstrategy.attachAnalyzer(sharpeRatioAnalyzer)

    # Attach the plotter to the strategy.

    plt = plotter.StrategyPlotter(tstrategy)

    # Include the SMA in the instrument's subplot to get it displayed along with the closing prices.
    # if other techinical indicators are used, add those data series here

    plt.getInstrumentSubplot(key).addDataSeries('SMA',
            tstrategy.getSMA())

    # Plot the strategy returns at each bar.

    plt.getOrCreateSubplot('net/daily returns'
                           ).addDataSeries('Net return',
            returnsAnalyzer.getReturns())
    plt.getOrCreateSubplot('cum. returns').addDataSeries('Cum. return',
            returnsAnalyzer.getCumulativeReturns())

    tstrategy.run()
    total_portfolio_value = total_portfolio_value \
        + tstrategy.getResult()
    print 'Final portfolio value for stock %s : $%.2f' % (key,
            tstrategy.getResult())
    print 'Anual return: %.2f %%' \
        % returnsAnalyzer.getCumulativeReturns()[-1]
    print 'Annualized Sharpe ratio: %.2f' \
        % sharpeRatioAnalyzer.getSharpeRatio(0, annualized=True)
    print 'Average daily return: %.2f %%' \
        % pytechstats.mean(returnsAnalyzer.getReturns())
    plt.plot()

print 'Total portfolio value: %s' % total_portfolio_value
